import {
  ClassType,
  PropertyType,
} from '@loafer/core/constants/types';

import * as ReflectConstants from '@loafer/core/constants/reflect';

import AnnotatedElement from './AnnotatedElement';
import Class from './Class';

export class Reflection {
  private _clazz: Class;

  public constructor(type: ClassType) {
    let parameterTypes: any[] = Reflect.getMetadata(ReflectConstants.DESIGN_PARAMTYPES, type.constructor);

    this._clazz = new Class(type, this.convertParamTypes(parameterTypes));
  }

  /**
   * getClass
   */
  public getClass(): Class {
    return this._clazz;
  }

  public addClassAnnotation(annotation: any): void {
    let parameterTypes: any[] = Reflect.getMetadata(ReflectConstants.DESIGN_PARAMTYPES, this._clazz.Type);

    const {valid, name} = this.validateAnnotation(annotation, this._clazz);
    if (!valid) {
      throw new Error(`Cannot apply @${name} decorator multiple times on Class[${this._clazz.Name}].`);
    }

    this._clazz.addAnnotation(annotation);
  }
  public addPropertyAnnotation(annotation: any, propertyKey: PropertyType): void {
    let propertyType: any = Reflect.getMetadata(ReflectConstants.DESIGN_TYPE, this._clazz.Type, propertyKey);

    let property = this._clazz.getProperty(propertyKey);
    if (undefined === property) {
      property = this._clazz.addProperty(propertyKey, propertyType.prototype);
    }

    const {valid, name} = this.validateAnnotation(annotation, property);
    if (!valid) {
      throw new Error(`Cannot apply @${name} decorator multiple times on Property[${this._clazz.Name}:${propertyKey}].`);
    }

    property.addAnnotation(annotation);
  }
  public addMethodAnnotation(annotation: any, propertyKey: PropertyType): void {
    let parameterTypes: any[] = Reflect.getMetadata(ReflectConstants.DESIGN_PARAMTYPES, this._clazz.Type, propertyKey);
    let returnType: any = Reflect.getMetadata(ReflectConstants.DESIGN_RETURNTYPE, this._clazz.Type, propertyKey);

    let method = this._clazz.getMethod(propertyKey);
    if (undefined === method) {
      method = this._clazz.addMethod(propertyKey, this.convertParamTypes(parameterTypes), returnType.prototype);
    }

    const {valid, name} = this.validateAnnotation(annotation, method);
    if (!valid) {
      throw new Error(`Cannot apply @${name} decorator multiple times on Method[${this._clazz.Name}:${propertyKey}].`);
    }

    method.addAnnotation(annotation);
  }
  public addParameterAnnotation(annotation: any, propertyKey: PropertyType, parameterIndex: number): void {
    let parameterTypes: any[] = Reflect.getMetadata(ReflectConstants.DESIGN_PARAMTYPES, this._clazz.Type, propertyKey);
    let returnType: any = Reflect.getMetadata(ReflectConstants.DESIGN_RETURNTYPE, this._clazz.Type, propertyKey);

    let method = this._clazz.getMethod(propertyKey);
    if (undefined === method) {
      method = this._clazz.addMethod(propertyKey, this.convertParamTypes(parameterTypes), returnType.prototype);
    }

    let parameter = method.getParameter(parameterIndex);

    const {valid, name} = this.validateAnnotation(annotation, parameter);
    if (!valid) {
throw new Error(`Cannot apply @${name} decorator multiple times on Parameter[${this._clazz.Name}:${propertyKey}(${parameterIndex}})].`);
    }

    parameter.addAnnotation(annotation);
  }

  private validateAnnotation(annotation: any, element: AnnotatedElement): {valid: boolean, name: string} {
    const annotationType: ClassType = Object.getPrototypeOf(annotation);
    const name: string = annotationType.constructor.name;

    if (element.hasAnnotation(annotationType)) {
      return {valid: false, name: name};
    }
    return {valid: true, name: name};
  }

  private convertParamTypes(parameterTypes: any[]): ClassType[] {
    let returnTypes: ClassType[] = [];
    parameterTypes.forEach((currentValue, index, array) => {
      returnTypes.push(currentValue.prototype);
    });
    return returnTypes;
  }
}

export default Reflection;
